"""JARVIS database models and utilities."""
from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorClient
from pytz import utc
from umongo.frameworks import MotorAsyncIOInstance

from jarvis_core.util import find

CLIENT = None
JARVISDB = None
CTC2DB = None
JARVIS_INST = MotorAsyncIOInstance()
CTC2_INST = MotorAsyncIOInstance()


def connect(
    username: str,
    password: str,
    port: int = 27017,
    testing: bool = False,
    host: str = None,
    hosts: list[str] = None,
    replicaset: str = None,
) -> None:
    """
    Connect to MongoDB.

    Args:
        host: Hostname/IP
        username: Username
        password: Password
        port: Port
    """
    global CLIENT, JARVISDB, CTC2DB, JARVIS_INST, CTC2_INST

    if not replicaset:
        CLIENT = AsyncIOMotorClient(
            host=host, username=username, password=password, port=port, tz_aware=True, tzinfo=utc
        )
    else:
        CLIENT = AsyncIOMotorClient(
            hosts, username=username, password=password, tz_aware=True, tzinfo=utc, replicaset=replicaset
        )
    JARVISDB = CLIENT.narvis if testing else CLIENT.jarvis
    CTC2DB = CLIENT.ctc2

    JARVIS_INST.set_db(JARVISDB)
    CTC2_INST.set_db(CTC2DB)


QUERY_OPS = ["ne", "lt", "lte", "gt", "gte", "not", "in", "nin", "mod", "all", "size"]
STRING_OPS = [
    "exact",
    "iexact",
    "contains",
    "icontains",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
    "wholeword",
    "iwholeword",
    "regex",
    "iregex" "match",
]
GEO_OPS = [
    "get_within",
    "geo_within_box",
    "geo_within_polygon",
    "geo_within_center",
    "geo_within_sphere",
    "geo_intersects",
    "near",
    "within_distance",
    "within_spherical_distance",
    "near_sphere",
    "within_box",
    "within_polygon",
    "max_distance",
    "min_distance",
]

ALL_OPS = QUERY_OPS + STRING_OPS + GEO_OPS


def q(**kwargs: dict) -> dict:
    """uMongo query wrapper."""  # noqa: D403
    query = {}
    for key, value in kwargs.items():
        if key == "_id":
            value = ObjectId(value)
        elif "__" in key:
            args = key.split("__")
            if not any(x in ALL_OPS for x in args):
                key = ".".join(args)
            else:
                idx = args.index(find(lambda x: x in ALL_OPS, args))
                key = ".".join(args[:idx])
                value = {f"${args[idx]}": value}
        query[key] = value
    return query
