"""JARVIS database models."""
import re
from datetime import datetime
from typing import Any, List

import marshmallow as ma
from umongo import Document, EmbeddedDocument, fields

from jarvis_core.db import CTC2_INST, JARVIS_INST
from jarvis_core.db.fields import RawField
from jarvis_core.db.models.actions import *
from jarvis_core.db.models.backups import *
from jarvis_core.db.models.mastodon import *
from jarvis_core.db.models.modlog import *
from jarvis_core.db.models.reddit import *
from jarvis_core.db.models.twitter import *
from jarvis_core.db.utils import get_now


@JARVIS_INST.register
class Autopurge(Document):
    guild: int = fields.IntegerField(required=True)
    channel: int = fields.IntegerField(required=True)
    delay: int = fields.IntegerField(default=30)
    admin: int = fields.IntegerField(required=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Autoreact(Document):
    guild: int = fields.IntegerField(required=True)
    channel: int = fields.IntegerField(required=True)
    reactions: List[str] = fields.ListField(fields.StringField())
    admin: int = fields.IntegerField(required=True)
    thread: bool = fields.BooleanField(default=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Config(Document):
    """Config database object."""

    key: str = fields.StringField(required=True)
    value: Any = RawField(required=True)


@JARVIS_INST.register
class Filter(Document):
    """Regex Filter database object."""

    def _validate_filters(value):
        for v in value:
            try:
                re.compile(v)
            except re.error:
                raise ValueError(f"Invalid regex: {v}")

    guild: int = fields.IntegerField(required=True)
    name: str = fields.StringField(required=True)
    filters: List[str] = fields.ListField(fields.StringField(), validate=[_validate_filters])


@CTC2_INST.register
class Guess(Document):
    """Guess database object."""

    correct: bool = fields.BooleanField(default=False)
    guess: str = fields.StringField(required=True)
    user: int = fields.IntegerField(required=True)


@JARVIS_INST.register
class Permission(EmbeddedDocument):
    """Embedded Permissions document."""

    id: int = fields.IntegerField(required=True)
    allow: int = fields.IntegerField(default=0)
    deny: int = fields.IntegerField(default=0)


@JARVIS_INST.register
class Lock(Document):
    """Lock database object."""

    active: bool = fields.BooleanField(default=True)
    admin: int = fields.IntegerField(required=True)
    channel: int = fields.IntegerField(required=True)
    duration: int = fields.IntegerField(default=10)
    guild: int = fields.IntegerField(required=True)
    reason: str = fields.StringField(required=True)
    original_perms: Permission = fields.EmbeddedField(Permission, required=False)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Lockdown(Document):
    """Lockdown database object."""

    active: bool = fields.BooleanField(default=True)
    admin: int = fields.IntegerField(required=True)
    duration: int = fields.IntegerField(default=10)
    guild: int = fields.IntegerField(required=True)
    reason: str = fields.StringField(required=True)
    original_perms: int = fields.IntegerField(required=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Event(Document):
    """Event Meetup Object."""

    user: int = fields.IntegerField(required=True)
    going: bool = fields.BooleanField(required=True)
    travel_method: str = fields.StringField()
    before_flight: str = fields.StringField()
    before_arrival_time: datetime = fields.AwareDateTimeField()
    before_departure_time: datetime = fields.AwareDateTimeField()
    after_flight: str = fields.StringField()
    after_arrival_time: datetime = fields.AwareDateTimeField()
    after_departure_time: datetime = fields.AwareDateTimeField()
    hotel: str = fields.StringField()
    event_name: str = fields.StringField()


@JARVIS_INST.register
class Phishlist(Document):
    """Phishing safelist."""

    url: str = fields.StringField(required=True)
    confirmed: bool = fields.BooleanField(default=False)
    valid: bool = fields.BooleanField(default=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Purge(Document):
    """Purge database object."""

    admin: int = fields.IntegerField(required=True)
    channel: int = fields.IntegerField(required=True)
    guild: int = fields.IntegerField(required=True)
    count: int = fields.IntegerField(default=10)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Reminder(Document):
    """Reminder database object."""

    active: bool = fields.BooleanField(default=True)
    user: int = fields.IntegerField(required=True)
    guild: int = fields.IntegerField(required=True)
    channel: int = fields.IntegerField(required=True)
    message: str = fields.StringField(required=True)
    remind_at: datetime = fields.AwareDateTimeField(required=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)
    private: bool = fields.BooleanField(default=False)


@JARVIS_INST.register
class Rolegiver(Document):
    """Rolegiver database object."""

    guild: int = fields.IntegerField(required=True)
    roles: List[int] = fields.ListField(fields.IntegerField())


@JARVIS_INST.register
class Bypass(EmbeddedDocument):
    """Roleping bypass embedded object."""

    users: List[int] = fields.ListField(fields.IntegerField())
    roles: List[int] = fields.ListField(fields.IntegerField())


@JARVIS_INST.register
class Roleping(Document):
    """Roleping database object."""

    active: bool = fields.BooleanField(default=True)
    role: int = fields.IntegerField(required=True)
    guild: int = fields.IntegerField(required=True)
    admin: int = fields.IntegerField(required=True)
    bypass: Bypass = fields.EmbeddedField(Bypass)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Setting(Document):
    """Setting database object."""

    guild: int = fields.IntegerField(required=True)
    setting: str = fields.StringField(required=True)
    value: Any = RawField()


@JARVIS_INST.register
class Star(Document):
    """Star database object."""

    active: bool = fields.BooleanField(default=True)
    index: int = fields.IntegerField(required=True)
    message: int = fields.IntegerField(required=True)
    channel: int = fields.IntegerField(required=True)
    starboard: int = fields.IntegerField(required=True)
    guild: int = fields.IntegerField(required=True)
    admin: int = fields.IntegerField(required=True)
    star: int = fields.IntegerField(required=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Starboard(Document):
    """Starboard database object."""

    channel: int = fields.IntegerField(required=True)
    guild: int = fields.IntegerField(required=True)
    admin: int = fields.IntegerField(required=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class Tag(Document):
    """Tag database object."""

    creator: int = fields.IntegerField(required=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)
    editor: int = fields.IntegerField()
    edited_at: datetime = fields.AwareDateTimeField()
    name: str = fields.StringField(required=True)
    content: str = fields.StringField(required=True)
    guild: int = fields.IntegerField(required=True)


@JARVIS_INST.register
class Temprole(Document):
    """Temporary role object."""

    guild: int = fields.IntegerField(required=True)
    user: int = fields.IntegerField(required=True)
    role: int = fields.IntegerField(required=True)
    admin: int = fields.IntegerField(required=True)
    expires_at: datetime = fields.AwareDateTimeField(required=True)
    created_at: datetime = fields.AwareDateTimeField(default=get_now)


@JARVIS_INST.register
class UserSetting(Document):
    """User Setting object."""

    user: int = fields.IntegerField(required=True)
    type: str = fields.StringField(required=True)
    setting: str = fields.StringField(required=True)
    value: Any = RawField()

    class Meta:
        collection_name = "usersetting"
