"""JARVIS quality of life utilities."""
import hashlib
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

from aiohttp import ClientSession

from jarvis_core.filters import url

DEFAULT_BLOCKSIZE = 8 * 1024 * 1024


class Singleton(object):
    REQUIRED = []
    OPTIONAL = {}

    def __new__(cls, *args: list, **kwargs: dict):
        """Create a new singleton."""
        inst = cls.__dict__.get("inst")
        if inst is not None:
            return inst

        inst = object.__new__(cls)
        inst.init(*args, **kwargs)
        inst._validate()
        cls.__inst__ = inst

        return inst

    def _validate(self) -> None:
        for key in self.REQUIRED:
            if not getattr(self, key, None):
                raise ValueError(f"Missing required key: {key}")

    def init(self, **kwargs: dict) -> None:
        """Initialize the object."""
        for key, value in kwargs.items():
            setattr(self, key, value)

        for key, value in self.OPTIONAL.items():
            if not getattr(self, key, None):
                setattr(self, key, value)


async def hash(
    data: str, method: Union[Callable, str] = hashlib.sha256, size: int = DEFAULT_BLOCKSIZE
) -> Tuple[str, int, str]:
    """
    Hash an object.

    Args:
        data: URL or text to hash
        method: Hash function/function name, default `hashlib.sha256`
        size: Blocksize, default `8 MB`

    Returns:
        Tuple of hash hex, data size, and content type

    Raises:
        Exception: Raise on status
    """
    # Get hashlib method
    if isinstance(method, str):
        method = getattr(hashlib, method, hashlib.sha256)
    method = method()

    # Hash non-URLs
    if not url.match(data):
        method.update(data.encode("UTF8"))
        return method.hexdigest(), len(data), "string"

    # Hash URL contents
    async with ClientSession() as session:
        resp = await session.get(data)
        resp.raise_for_status()
        content_type = resp.content_type

        # While we could use resp.content_length, that sometimes returns None,
        # so we manually count the size while looping.
        data_len = 0
        block_count = 0

        async for block in resp.content.iter_chunked(n=size):
            data_len += len(block)
            block_count += 1
            method.update(block)

        return method.hexdigest(), data_len, content_type


def convert_bytesize(b: int) -> str:
    """Convert bytes amount to human readable."""
    if not b:
        return "??? B"
    b = float(b)
    sizes = ["B", "KB", "MB", "GB", "TB", "PB"]
    size = 0
    while b >= 1024 and size < len(sizes) - 1:
        b = b / 1024
        size += 1
    return "{:0.3f} {}".format(b, sizes[size])


def unconvert_bytesize(size: int, ending: str) -> int:
    """Convert human readable to bytes."""
    ending = ending.upper()
    sizes = ["B", "KB", "MB", "GB", "TB", "PB"]
    if ending == "B":
        return size
    # Rounding is only because bytes cannot be partial
    return round(size * (1024 ** sizes.index(ending)))


def find(predicate: Callable, sequence: Iterable) -> Optional[Any]:
    """
    Find the first element in a sequence that matches the predicate.

    ??? Hint "Example Usage:"
        ```python
        member = find(lambda m: m.name == "UserName", guild.members)
        ```
    Args:
        predicate: A callable that returns a boolean value
        sequence: A sequence to be searched

    Returns:
        A match if found, otherwise None

    """
    for el in sequence:
        if predicate(el):
            return el
    return None


def find_all(predicate: Callable, sequence: Iterable) -> List[Any]:
    """
    Find all elements in a sequence that match the predicate.

    ??? Hint "Example Usage:"
        ```python
        members = find_all(lambda m: m.name == "UserName", guild.members)
        ```
    Args:
        predicate: A callable that returns a boolean value
        sequence: A sequence to be searched

    Returns:
        A list of matches

    """
    return [el for el in sequence if predicate(el)]


def get(sequence: Iterable, **kwargs: Any) -> Optional[Any]:
    """
    Find the first element in a sequence that matches all attrs.

    ??? Hint "Example Usage:"
        ```python
        channel = get(guild.channels, nsfw=False, category="General")
        ```

    Args:
        sequence: A sequence to be searched
        kwargs: Keyword arguments to search the sequence for

    Returns:
        A match if found, otherwise None
    """
    if not kwargs:
        return sequence[0]

    for el in sequence:
        if any(not hasattr(el, attr) for attr in kwargs.keys()):
            continue
        if all(getattr(el, attr) == value for attr, value in kwargs.items()):
            return el
    return None


def get_all(sequence: Iterable, **kwargs: Any) -> List[Any]:
    """
    Find all elements in a sequence that match all attrs.

    ??? Hint "Example Usage:"
        ```python
        channels = get_all(guild.channels, nsfw=False, category="General")
        ```

    Args:
        sequence: A sequence to be searched
        kwargs: Keyword arguments to search the sequence for

    Returns:
        A list of matches
    """
    if not kwargs:
        return sequence

    matches = []
    for el in sequence:
        if any(not hasattr(el, attr) for attr in kwargs.keys()):
            continue
        if all(getattr(el, attr) == value for attr, value in kwargs.items()):
            matches.append(el)
    return matches
